/* bfbtester -- Brute Force Binary Tester
 *
 * Copyright 2000	Mike Heffner <spock@techfour.net>
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#include <sys/types.h>
#include <dirent.h>

#include "datatypes.h"
#include "data_stuff.h"
#include "main.h"
#include "exec.h"
#include "breaker.h"
#include "utils.h"
#include "tempfiles.h"

#define MAX_SIMULTANEOUS_EXECS	250

static void
usage(char *progname);

static void
run_breaker(void);

static void
load_execs(char *location, ExecEntryStatus status);

THREAD_DATA	*q;
static int	flags=0;
static int	t_on = FALSE;		/* tempfile monitoring */
char		*run_location;
char		*rejects = NULL;
FILE		*fp_msg = NULL;

extern int	optind;
extern char	*optarg;

int
main( int ac, char ** av )
{
	int ch;
	char *progname = av[0];

	while ((ch = getopt(ac, av, "ad:ehmo:r:stv")) != -1)
		switch (ch) {
		case 'a':
			flags |= FLAGS_ALL;
			break;
		case 'd':
			/* backwards compat */
			break;
		case 'e':
			flags |= FLAGS_ENV_TEST;
			break;
		case 'h':
			usage(progname);
			break;
		case 'm':
			flags |= FLAGS_MULTIPLE_ARG_TEST;
			break;
		case 'o':
			if ((fp_msg = fopen(optarg, "w")) == NULL) {
				printf("Can't open output file: %s\n", optarg);
				return -1;
			}
			break;
		case 'r':
			rejects = optarg;
			break;
		case 's':
			flags |= FLAGS_SINGLE_ARG_TEST;
			break;
		case 't':
			t_on = TRUE;
			break;
		case 'v':
			printf("%s v%s\n", IDSTR, VERSTR);
			return 0;
			/* NO FALL-THROUGH */
		default:
			usage(progname);
			/* NOTREACHED */
		}

	ac -= optind;
	av += optind;

	if (fp_msg == NULL)
		fp_msg = stdout;
	if (!flags || (run_location = *av) == NULL)
		usage(progname);

	q = xcalloc(1, sizeof(THREAD_DATA));
	q->max_execs = MAX_SIMULTANEOUS_EXECS;
	q->mut = xmalloc(sizeof(pthread_mutex_t));

	if (pthread_mutex_init(q->mut, NULL) != 0)
		bfbt_message(MSG_TYPE_ERRNO_ABORT, "mutex init");

	if (t_on)
		tempfile_start(q);

	load_execs(run_location, EXEC_ENTRY_STATUS_TEST);
	run_breaker();

	return 0;

}

static void
usage(char *progname)
{
	printf("%s [-htv] [-r rejects] [-o out-file] -asme BinOrDir\n"
		   "Must specify one or more of the following tests:\n"
		   "    -s	Single Argument Test\n"
		   "    -m	Multiple Argument Test\n"
		   "    -e	Environment Variable Test\n"
		   "    -a	Selects all tests\n"
		   "Options:\n"
		   "    -h          Print this message\n"
		   "    -t          Enable tempfile monitoring\n"
		   "    -v          Print version string\n"
		   "    -r rejects  Comma separated list of binaries to skip\n"
		   "    -o out-file Output to out-file rather than stdout\n"
		   "BinOrDir    Specific Binary or Directory of binaries to test\n",
		   progname);
	exit(2);
}

static void
checknadd_exec(char *filename, ExecEntryStatus status,
			   int test_flags)
{
	/* Test if exec is to be rejected */
	if (rejects) {
		char *ptr, *ptr2;

		if((ptr=strstr(rejects, filename)))
			if(ptr == rejects || *(ptr-1) == ' ' || *(ptr-1) == ',')
				if(*(ptr+strlen(filename)) == '\0' || *(ptr+strlen(filename)) == ' ' ||
				   *(ptr+strlen(filename)) == ',')
					return;
		if( (ptr2=strrchr(filename, '/')) && (ptr=strstr(rejects, ++ptr2)) )
			if(ptr == rejects || *(ptr-1) == ' ' || *(ptr-1) == ',')
				if(*(ptr+strlen(ptr2)) == '\0' || *(ptr+strlen(ptr2)) == ' ' ||
				   *(ptr+strlen(ptr2)) == ',')
					return;
	}
	/* If we get to here we are safe to add it */
	ExecEntry_add(filename, status, flags);
}

static void
load_execs(char *location, ExecEntryStatus status)
{
	if (location[0] == '/') {
		struct stat sb;
		if (stat(location, &sb) != 0) {
			bfbt_message(MSG_TYPE_ERRNO_WARN, "Can't stat %s", location);
			return;
		}
		if (S_ISDIR(sb.st_mode)) {		/* Walk directory */
				DIR *dir;
				struct dirent entry, *result;
				int ret;
				char fname[2 * FILENAME_MAX];
				
				if ((dir = opendir(location)) == NULL) {
					bfbt_message(MSG_TYPE_ERROR_WARN, "Can't open dir %s", 
								 location);
					return;
				}
				ret = readdir_r(dir, &entry, &result);
				while (ret == 0 && result != NULL) {
					if (snprintf(fname, sizeof(fname), "%s%s%s", location,
								 location[strlen(location)-1] == '/' ?
								 "" : "/", result->d_name) < sizeof(fname))
						checknadd_exec(fname, status, flags);
					ret = readdir_r(dir, &entry, &result);
				}
		} else							/* Treat as an executable */
			checknadd_exec(location, status, flags);
	} else {
		char *filename = findExec(location);
		if (filename)
			checknadd_exec(filename, status, flags);
		free(filename);
	}
}

static void
run_breaker(void)
{
	if (!ExecEntry_isempty()) {
		breaker_start();
		breaker_wait();
	}
}
