/* bfbtester -- Brute Force Binary Tester
 *
 * Copyright 2000	Mike Heffner <spock@techfour.net>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/wait.h>


#include "datatypes.h"
#include "data_stuff.h"
#include "utils.h"

#include "bfbt_queue.h"

/* E X E C   E N T R Y  */

DLIST_HEAD(exec_entry_head, _exec_entry) ExecEntry_Head;

static void
ExecEntry_constructor(void)				__attribute__ ((constructor));

static void
ExecEntry_constructor(void)
{
	DLIST_INIT(&ExecEntry_Head);
}

static inline void
ExecEntry_init(ExecEntry *entry, struct stat sb, char *fullname, 
			   ExecEntryStatus status, int test_flags)
{
	entry->fullname = strdup(fullname);
	entry->filestat = sb;
	entry->status = status;
	entry->test_flags = test_flags;
	entry->testtime.tv_sec = 0;
	entry->testtime.tv_usec = 0;
}

static void
ExecEntry_destroy(ExecEntry *entry)
{
	assert(entry != NULL);
	if (entry->fullname)	free(entry->fullname);
/*
	if (entry->crash_reports) {
		if (entry->crash_destructor)
			(*entry->crash_destructor)(entry->crash_reports);
		else
			CrashReports_destroychain(entry->crash_reports);
	}
*/
#if 0
	if (entry->no_tempfiles) {
		if (entry->tempfiles_destructor)
			(*entry->tempfiles_destructor)(entry->tempfiles);
		else
			(*entry->tempfiles_destroychain)(entry->tempfiles);
	}
#endif
	free(entry);
}

static inline int
ExecEntry_compare(ExecEntry *e1, ExecEntry *e2)
{
	if (e1->filestat.st_ino == e2->filestat.st_ino)
		return TRUE;
	else
		return FALSE;
}

ExecEntry *
ExecEntry_add(char *name, ExecEntryStatus status, int test_flags)
{
	ExecEntry *new;
	struct stat sb;

	assert(name != NULL);

	if (stat(name, &sb) != 0 || !S_ISREG(sb.st_mode)
		|| exec_ispermok(name) == FALSE)
		return NULL;

	new = xmalloc(sizeof(ExecEntry));
	ExecEntry_init(new, sb, name, status, test_flags);
	DLIST_INSERT_HEAD(&ExecEntry_Head, new, dlist_next);
	return new;
}

void
ExecEntry_subtract(ExecEntry *entry)
{
	assert(entry != NULL);
	DLIST_REMOVE(&ExecEntry_Head, entry, dlist_next);
	ExecEntry_destroy(entry);
}

ExecEntry *
ExecEntry_getnext(ExecEntry **entry)
{
	DLIST_GETNEXT(&ExecEntry_Head, entry, dlist_next);
	return *entry;
}

ExecEntry *
ExecEntry_findbyname(char *name)
{
	ExecEntry *entry = NULL;
	
	entry = ExecEntry_getnext(&entry);
	while (entry && strcmp(entry->fullname, name) != 0)
		entry = ExecEntry_getnext(&entry);
	return entry;
}

void
ExecEntry_release(ExecEntry *entry)
{
	DLIST_RELEASE(entry, dlist_next);
}

int
ExecEntry_isempty(void)
{
	return DLIST_ISEMPTY(&ExecEntry_Head);
}

/* T E S T   E N T R Y */

TAILQ_HEAD(test_entry_head, _test_entry) TestEntry_Head;

static void
TestEntry_constructor(void)			__attribute__  ((constructor));

static void
TestEntry_constructor(void)
{
	TAILQ_INIT(&TestEntry_Head);
}

extern char ** environ;

static inline void
TestEntry_init(TestEntry *entry, ExecEntry *parent, char **args, char **env)
{
	assert(entry != NULL);
	entry->pid = -1;
	entry->status = 0;
	if ((entry->progname = strdup(parent->fullname)) == NULL)
		bfbt_message(MSG_TYPE_ERRNO_ABORT, "Out of memory");
	entry->args = args;
	entry->args_str = args_to_str(entry->args);
	if (env) {
		entry->env = env_create(env);
		entry->env_flag = FALSE;
		entry->env_str = env_to_str(env);
	} else {
		entry->env = environ;
		entry->env_flag = TRUE;
		entry->env_str = NULL;
	}
	entry->parent = parent;
}

void
TestEntry_destroy(TestEntry *entry)
{
	assert(entry != NULL);
	if (entry->progname)	free(entry->progname);
	if (entry->args_str)	free(entry->args_str);
	if (entry->env && entry->env_flag == FALSE)	free(entry->env);
	if (entry->env_str)		free(entry->env_str);
	free(entry);
}

TestEntry *
TestEntry_create(ExecEntry *parent, char **args, char **env)
{
	TestEntry *new;

	new = xmalloc(sizeof(TestEntry));
	TestEntry_init(new, parent, args, env);

	return new;
}

void
TestEntry_insert(TestEntry *entry)
{
	assert(entry != NULL && entry->pid != -1);

	if (TAILQ_EMPTY(&TestEntry_Head) 
		|| entry->pid < TAILQ_FIRST(&TestEntry_Head)->pid)
		TAILQ_INSERT_HEAD(&TestEntry_Head, entry, links);
	else if (entry->pid > TAILQ_LAST(&TestEntry_Head, test_entry_head)->pid)
		TAILQ_INSERT_TAIL(&TestEntry_Head, entry, links);
	else {
		TestEntry *temp = TAILQ_FIRST(&TestEntry_Head);
		while (entry->pid > temp->pid)
			temp = TAILQ_NEXT(temp, links);
		if (entry->pid == temp->pid)
			bfbt_message(MSG_TYPE_ERROR_ABORT, "pid collision %s = %s",
						 entry->parent->fullname, temp->parent->fullname);
		TAILQ_INSERT_BEFORE(temp, entry, links);
	}
}

void
TestEntry_remove(TestEntry *entry)
{
	assert(entry != NULL && entry->pid != -1 && !TAILQ_EMPTY(&TestEntry_Head));

	TAILQ_REMOVE(&TestEntry_Head, entry, links);
}

TestEntry *
TestEntry_getfirst(void)
{
	return TAILQ_FIRST(&TestEntry_Head);
}

TestEntry *
TestEntry_getnext(TestEntry *entry)
{
	return TAILQ_NEXT(entry, links);
}

/* C R A S H   R E P O R T S */

#include <signal.h>

SLIST_HEAD(crashreport_slist_head, _crash_report) CrashReport_Head;

static void
CrashReport_constructor(void)				__attribute__ ((constructor));

static void
CrashReport_constructor(void)
{
	SLIST_INIT(&CrashReport_Head);
}

static void
CrashReport_init(CrashReport *report, TestEntry *test_entry)
{
	report->progname = strdup(test_entry->progname);
	report->args_str = (test_entry->args_str ? strdup(test_entry->args_str)
						: NULL);
	report->env_str = (test_entry->env_str ? strdup(test_entry->env_str)
					   : NULL);
	report->signal	= (int)WTERMSIG(test_entry->status);
	report->didCore = (int)WCOREDUMP(test_entry->status);
}

static void
CrashReport_destroy(CrashReport *report)
{
	if (report->progname)	free(report->progname);
	if (report->args_str)	free(report->args_str);
	if (report->env_str)	free(report->env_str);
	free(report);
}

CrashReport *
CrashReport_add(TestEntry *test_entry)
{
	CrashReport	*new;

	assert(test_entry != NULL);
	new = xmalloc(sizeof(CrashReport));
	CrashReport_init(new, test_entry);
	SLIST_INSERT_HEAD(&CrashReport_Head, new, link);
	return new;
}

void
CrashReport_subtract(CrashReport *report)
{
	assert(report != NULL);
	SLIST_REMOVE(&CrashReport_Head, report, _crash_report, link);
	CrashReport_destroy(report);
}

void
CrashReport_log(CrashReport *report)
{
	bfbt_message(MSG_TYPE_MESSAGE,
				 "*** Crash <%s> ***\n"
				 "args:		%s\n"
				 "envs:		%s\n"
				 "Signal:		%d ( %s )\n"
				 "Core?		%s\n",
				 report->progname,
				 report->args_str, report->env_str,
				 report->signal, strsignal(report->signal),
				 report->didCore ? "Yes" : "No");
}
