/* bfbtester -- Brute Force Binary Tester
 *
 * Copyright 2000	Mike Heffner <spock@techfour.net>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>

#include "datatypes.h"
#include "data_stuff.h"
#include "exec.h"
#include "utils.h"

extern	THREAD_DATA	*q;

int
execute_program(ExecEntry *exec_entry, char ** args, char ** env)
{
	TestEntry *test_entry;

	if( args == NULL )
		return -1;
	if( q->current_execs >= q->max_execs ){
		bfbt_message(MSG_TYPE_DEBUG_L1, "Hit maximum execs");
		return -1;
	}
	test_entry = TestEntry_create(exec_entry, args, env);

	switch( test_entry->pid = fork() ){
	case -1:

		bfbt_message(MSG_TYPE_ERROR_WARN, "Couldn't fork process");
		TestEntry_destroy(test_entry);
		return (1);

	case 0:
		
	{
		pid_t	pgid = getpid();
		setpgid(pgid, pgid);
	}
	
		close(0);
		close(1);
		close(2);

		execve(test_entry->args[0], test_entry->args, test_entry->env);

		/* if we get here then the execve() didn't work 
		 * NOTE: The error message will not print if we are
		 * outputting to stdout/stderr
		 */
		bfbt_message(MSG_TYPE_ERRNO_WARN, "execve '%s'", test_entry->args[0]);
		_exit(2);

	default:

		gettimeofday(&(test_entry->tm), NULL);

		pthread_mutex_lock(q->mut);
		TestEntry_insert(test_entry);
		q->current_execs++;
		pthread_mutex_unlock(q->mut);

	}
	return (0);
}

static void
exec_kill(pid_t pid)
{
	kill(-(pid), SIGKILL);
}

static int
exec_IsDone(TestEntry *entry)
{
	struct timeval	tm;
#define	EXEC_STALE		20	/* seconds */

	if( waitpid(entry->pid, &(entry->status), WNOHANG) > 0 )
		return TRUE;
	gettimeofday(&tm, NULL);
	if( tm.tv_sec - entry->tm.tv_sec > EXEC_STALE )
		bfbt_message(MSG_TYPE_ERROR_WARN, "Stale: %ld secs '%s'\n",
					 tm.tv_sec - entry->tm.tv_sec, entry->args[0]);
	if( tm.tv_sec - entry->tm.tv_sec > EXEC_TIMEOUT ){
		exec_kill(entry->pid);
		if( waitpid(entry->pid, &(entry->status), WNOHANG) > 0 )
			return TRUE;
	}
	return FALSE;
}

static void
exec_process_close(TestEntry *entry)
{
	q->current_execs--;
	if( WIFEXITED(entry->status) )
		; /* nothing bad happened to it */
	else if( WIFSIGNALED(entry->status) ){
		/* we ignore SIGKILL and SIGPIPE */
		if( WTERMSIG(entry->status) == SIGKILL
			|| WTERMSIG(entry->status) == SIGPIPE){
			return;
		}
		{
			/* we crashed -- record */
			CrashReport * report = CrashReport_add(entry);
			CrashReport_log(report);
		}
	} else
		bfbt_message(MSG_TYPE_ERROR_WARN, 
					 "%s ended in unknown condition, status=%d",
					 entry->args[0], entry->status);
}


void
exec_poll(void)
{
	TestEntry *temp, *dead;
	
	pthread_mutex_lock(q->mut);
	for (temp = TestEntry_getfirst(); temp ;){
		if (exec_IsDone(temp)){
			exec_process_close(temp);
			dead = temp;
			temp = TestEntry_getnext(temp);
			TestEntry_remove(dead);
			TestEntry_destroy(dead);
		} else
			temp = TestEntry_getnext(temp);
	}
	pthread_mutex_unlock(q->mut);
}

